import extract from '@emmetio/extract-abbreviation';
import { expand } from '@emmetio/expand-abbreviation';

/**
 * Returns token used for single indentation in given editor
 * @param  {CodeMirror} editor
 * @return {String}
 */
function getIndentation(editor) {
	if (!editor.getOption('indentWithTabs')) {
		return repeatString(' ', editor.getOption('indentUnit'));
	}

	return '\t';
}

/**
 * Normalizes text according to given CodeMirror instance indentation
 * preferences
 * @param  {String}     text
 * @param  {CodeMirror} editor
 * @param  {String}     [indentation] Applies `indentText()` with given argument,
 *                                    if provided
 * @return {String}
 */
function normalizeText(editor, text, indentation) {
	let lines = splitByLines(text);
	const indent = getIndentation(editor);

	if (indent !== '\t') {
		lines = lines.map(line => line.replace(/^\t+/,
			tabs => repeatString(indent, tabs.length)));
	}

	if (indentation) {
		lines = lines.map((line, i) => i ? indentation + line : line);
	}

	return lines.join('\n');
}

/**
 * Indents each line, except first one, in given text
 * @param  {String} text
 * @param  {String} indentation
 * @return {String}
 */


/**
 * Splits given text by lines
 * @param  {String} text
 * @return {String[]} Lines of text
 */
function splitByLines(text) {
	return Array.isArray(text) ? text : text.split(/\r\n|\r|\n/g);
}

function repeatString(str, count) {
	let result = '';
	while (0 < count--) {
		result += str;
	}

	return result;
}

const cursorMark = '[[::emmet-cursor::]]';

/**
 * Expand abbreviation command
 * @param  {CodeMirror} editor
 */
var emmetExpandAbbreviation = function(editor) {
	const cursor = editor.getCursor();
	const line = editor.getLine(cursor.line);

	// Extract Emmet abbreviation from current line with “look-ahead”,
	// e.g. capture auto-closed braces like `)`, `}` or `]` right after current
	// cursor position
	const extracted = extract(line, cursor.ch, true);
	if (!extracted) {
		// No valid abbreviation, abort key handling
		return editor.constructor.Pass;
	}

	let cursorMarked = false;
	let newSelectionSize = 0;
	const expandOpt = Object.assign({
		// CodeMirror doesn’t support snippets with tab-stops natively so we have
		// to mark first output with a special token so we can find it later
		// to properly plant cursor into new position
		field(index, placeholder = '') {
			if (!cursorMarked) {
				cursorMarked = true;
				newSelectionSize = placeholder.length;
				placeholder = cursorMark + placeholder;
			}

			return placeholder;
		}
	}, getExpandOptions(editor, cursor));

	const matchIndent = line.match(/^\s+/);
	let expanded;
	try {
		expanded = normalizeText(editor, expand(extracted.abbreviation, expandOpt), matchIndent && matchIndent[0]);
	} catch (err) {
		// Invalid abbreviation, abort key handling
		console.warn(err);
		return editor.constructor.Pass;
	}

	if (extracted.abbreviation === expanded) {
		// Edge case: expanded result is the same as abbreviation.
		// May happen in CSS and indent-based syntaxes like Slim, Pug
		return editor.constructor.Pass;
	}

	let newCursorPos = expanded.length;

	if (cursorMarked) {
		// Remove cursor stub and re-position cursor
		newCursorPos = expanded.indexOf(cursorMark);
		expanded = expanded.slice(0, newCursorPos) + expanded.slice(newCursorPos + cursorMark.length);
	}

	// Replace abbreviation with expanded result
	const rangeStart = {
		line: cursor.line,
		ch: extracted.location
	};
	const rangeEnd = {
		line: cursor.line,
		ch: extracted.location + extracted.abbreviation.length
	};
	editor.replaceRange(expanded, rangeStart, rangeEnd);

	// Position cursor
	const startIx = editor.indexFromPos(rangeStart);
	const newCursor = editor.posFromIndex(newCursorPos + startIx);
	if (newSelectionSize) {
		editor.setSelection(newCursor, {
			line: newCursor.line,
			ch: newCursor.ch + newSelectionSize
		});
	} else {
		editor.setCursor(newCursor);
	}
};

/**
 * Returns options object for syntax from given editor. In most cases, it detects
 * XML-style syntax (HTML, XML, XHTML) and returns options configured for proper
 * output
 * @param  {CodeMirror} editor
 * @param  {Point}      [pos]  Point in editor where syntax should be detected.
 *                             Uses `editor.getCursor()` in not given
 * @return {Object}
 */
function getExpandOptions(editor, pos) {
	const mode = editor.getModeAt(pos || editor.getCursor());
	const emmetOpt = editor.getOption('emmet') || {};
	let syntax = emmetOpt.syntax || mode.name;
	let profile = emmetOpt.profile;

	if (syntax === 'xml') {
		syntax = 'html';
		profile = Object.assign({ selfClosingStyle: mode.configuration }, profile);
	}

	return Object.assign({ syntax, profile }, emmetOpt);
}

/**
 * Inserts formatted line tag between tags
 * @param  {CodeMirror} editor
 */
var emmetInsertLineBreak = function(editor) {
	const cursor = editor.getCursor();
	const mode = editor.getModeAt(cursor);

	if (mode.name === 'xml') {
		const next = Object.assign({}, cursor, { ch: cursor.ch + 1 });
		const left = editor.getTokenAt(cursor);
		const right = editor.getTokenAt(Object.assign({}, cursor, { ch: cursor.ch + 1 }));

		if (left.type === 'tag bracket' && left.string === '>'
			&& right.type === 'tag bracket' && right.string === '</') {
				const matchIndent = editor.getLine(cursor.line).match(/^\s+/);
				const curIndent = matchIndent ? matchIndent[0] : '';
				const indent = getIndentation(editor);

				// Insert formatted line break
				const before = `\n${curIndent}${indent}`;
				const after = `\n${curIndent}`;
				editor.replaceRange(before + after, cursor, cursor);

				// Position cursor
				const startIx = editor.indexFromPos(cursor);
				const newCursor = editor.posFromIndex(startIx + before.length);
				editor.setCursor(newCursor);

				return;
			}
	}

	return editor.constructor.Pass;
};

const globalCommands = { emmetExpandAbbreviation, emmetInsertLineBreak };

/**
 * Adds Emmet support to given CodeMirror editor instance
 * @param  {CodeMirror} editor
 * @return {Function} A function that, when called, completely removes Emmet
 * support from editor instance
 */
var plugin = function(editor, options) {
	options = options || {};
	const keymap = options.keymap || {
		Tab: 'emmetExpandAbbreviation',
		'Enter': 'emmetInsertLineBreak'
	};

	registerCommands(editor, globalCommands);
	editor.addKeyMap(keymap);
	editor.setOption('emmet', options);

	return () => {
		editor.setOption('emmet', null);
		editor.removeKeyMap(keymap);
	};
};

/**
 * Registers editor commands in given editor, if required
 * @param  {CodeMirror} editor   Editor instance
 * @param  {Object}     commands Commands map
 */
function registerCommands(editor, commands) {
	const CodeMirror = editor.constructor;

	Object.keys(commands).forEach(name => {
		if (!CodeMirror.commands[name]) {
			CodeMirror.commands[name] = commands[name];
		}
	});
}

export default plugin;
